// Load modules

var Handlebars = require("handlebars");
var Hoek = require("hoek");
var Url = require("url");
var Ws = require("ws");

// Declare internals

var internals = {};

internals.defaults = {
    endpoint: "/manager/console",
    queryKey: "debug",
    template: "index",
    appHost: "127.0.0.1",
    appPort: "9000",
    publicHost: "127.0.0.1",
    publicPort: "9000"
};

exports.register = function (plugin, options, next) {

    var settings = Hoek.applyToDefaults(internals.defaults, options);

    plugin.views({
        engines: {html: {module: Handlebars}},
        path: __dirname + "/../templates"
    });

    plugin.route({
        method: "GET",
        path: settings.endpoint + "/{file*2}",
        config: {
            auth: false,
            handler: {directory: {path: __dirname + "/../public", listing: false, index: false}}
        }
    });

    var subscribers = {};
    var server = new plugin.hapi.Server(settings.appHost, settings.appPort);

    server.start(function () {

        var ws = new Ws.Server({server: server.listener});
        ws.on("connection", function (socket) {

            socket.on("message", function (message) {

                subscribers[message] = subscribers[message] || [];
                subscribers[message].push(socket);
            });
        });

        var context = {
            endpoint: settings.endpoint,
            host: settings.publicHost,
            port: settings.publicPort
        };

        plugin.route({
            method: "GET",
            path: settings.endpoint,
            config: {                      // In case defaults are set otherwise
                handler: function (request, reply) {
                    reply.view(settings.template, context);
                }
            }
        });

        plugin.ext("onRequest", function (request, extNext) {

            var key = settings.queryKey;
            if (request.query[key]) {
                request.plugins.tv = {debugId: request.query[key]};

                delete request.query[key];
                delete request.url.search;
                delete request.url.query[key];

                request.setUrl(Url.format(request.url));
            }

            return extNext();
        });

        plugin.events.on("request", function (request, report) {

            var transmit = function (key) {

                var subs = subscribers[key];
                if (!subs) {
                    return;
                }

                if (report.data && report.data.url) {
                    var pathSplitted = report.data.url.split("/");
                    if (pathSplitted[1] && (pathSplitted[1] == "public" || pathSplitted[1] == "mechanic")) {
                        return;
                    }
                }
                else {
                    return;
                }

                var valid = [];
                for (var i = 0, il = subs.length; i < il; ++i) {
                    try {
                        if (subs[i].readyState === Ws.OPEN) {
                            subs[i].send(JSON.stringify(report, null, 4));
                            valid.push(subs[i]);
                        }
                    }
                    catch (err) {
                    }
                }

                subscribers[key] = valid;                         // Removes subscriber on any send error
            };

            var session = request.plugins.tv && request.plugins.tv.debugId;
            if (session) {
                transmit(session);
            }

            transmit("*");
        });

        return next();
    });
};

exports.register.attributes = {
    pkg: require("../package.json")
};